searchData={"items":[{"type":"module","title":"Instructor","doc":"Instructor.ex is a spiritual port of the great [Instructor Python Library](https://github.com/jxnl/instructor) by [@jxnlco](https://twitter.com/jxnlco), check out his [talk](https://www.youtube.com/watch?v=yj-wSRJwrrc) on youtube.\nThis library brings structured prompting to LLMs. Instead of receiving text as output, Instructor will coax the LLM to output valid JSON that maps directly to the provided Ecto schema.\nIf the LLM fails to do so, or provides values that do not pass your validations, it will provide you utilities to automatically retry with the LLM to correct errors.\nBy default it's designed to be used with the [OpenAI API](https://platform.openai.com/docs/api-reference/chat-completions/create), however it provides an extendable adapter behavior to work with [ggerganov/llama.cpp](https://github.com/ggerganov/llama.cpp) and [Bumblebee (Coming Soon!)](https://github.com/elixir-nx/bumblebee).\n\nAt its simplest, usage is pretty straightforward,\n\n```elixir\ndefmodule SpamPrediction do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @doc \"\"\"","ref":"Instructor.html"},{"type":"module","title":"Field Descriptions: - Instructor","doc":"- class: Whether or not the email is spam\n  - reason: A short, less than 10 word rationalization for the classification\n  - score: A confidence score between 0.0 and 1.0 for the classification\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl true\n  def validate_changeset(changeset) do\n    changeset\n    |> Ecto.Changeset.validate_number(:score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n\nis_spam? = fn text ->\n  Instructor.chat_completion(\n    model: \"gpt-3.5-turbo\",\n    response_model: SpamPredicition,\n    max_retries: 3,\n    messages: [\n      %{\n        role: \"user\",\n        content: \"\"\"\n        You purpose is to classify customer support emails as either spam or not.\n        This is for a clothing retailer business.\n        They sell all types of clothing.\n\n        Classify the following email: #{text}\n        \"\"\"\n      }\n    ]\n  )\nend\n\nis_spam?.(\"Hello I am a Nigerian prince and I would like to send you money\")\n\n# => {:ok, %SpamPrediction{class: :spam, reason: \"Nigerian prince email scam\", score: 0.98}}\n```\n\nSimply create an ecto schema, optionally provide a `@doc` to the schema definition which we pass down to the LLM, then make a call to `Instructor.chat_completion/1` with context about the task you'd like the LLM to complete.\nYou can also provide a `validate_changeset/1` function via the `use Instructor.Validator` which will provide a set of code level ecto changeset validations. You can use this in conjunction with `max_retries: 3` to automatically, iteratively go back and forth with the LLM up to `n` times with any validation errors so that it has a chance to fix them.\n\n**Curious to learn more? Unsure of how you'd use this? Check out our extensive set of [tutorials](https://hexdocs.pm/instructor/tutorial.html)**\n\n* [Tutorial - Basic Usage & Features](https://hexdocs.pm/instructor/tutorial.html)\n* [Text Classification](https://hexdocs.pm/instructor/text-classification.html)\n* [Extracting Action Items from Meeting Transcriptions](https://hexdocs.pm/instructor/extract-action-items-from-meeting-transcripts.html)\n* [Extracting Explorer.DataFrames from text](https://hexdocs.pm/instructor/text-to-dataframes.html)","ref":"Instructor.html#module-field-descriptions"},{"type":"module","title":"Configuration - Instructor","doc":"To configure the default OpenAI adapter you can set the configuration,\n\n```elixir\nconfig :openai, api_key: \"sk-........\"\nconfig :openai, http_options: [recv_timeout: 10 * 60 * 1000]\n```\n\nTo use a local LLM, you can install and run [llama.cpp server](https://github.com/ggerganov/llama.cpp/blob/master/examples/server/README.md) and tell instructor to use it,\n\n```elixir\nconfig :instructor, adapter: Instructor.Adapters.Llamacpp\n```","ref":"Instructor.html#module-configuration"},{"type":"function","title":"Instructor.cast_all/2","doc":"","ref":"Instructor.html#cast_all/2"},{"type":"function","title":"Instructor.chat_completion/1","doc":"Create a new chat completion for the provided messages and parameters.\n\nThe parameters are passed directly to the LLM adapter.\nBy default they shadow the OpenAI API parameters.\nFor more information on the parameters, see the [OpenAI API docs](https://platform.openai.com/docs/api-reference/chat-completions/create).\n\nAdditionally, the following parameters are supported:\n\n  * `:response_model` - The Ecto schema to validate the response against.\n  * `:max_retries` - The maximum number of times to retry the LLM call if it fails, or does not pass validations.\n                     (defaults to `0`)","ref":"Instructor.html#chat_completion/1"},{"type":"function","title":"Examples - Instructor.chat_completion/1","doc":"iex> Instructor.chat_completion(%{\n  ...>   model: \"gpt-3.5-turbo\",\n  ...>   response_model: Instructor.Demos.SpamPrediction,\n  ...>   messages: [\n  ...>     %{\n  ...>       role: \"user\",\n  ...>       content: \"Classify the following text: Hello, I am a Nigerian prince and I would like to give you $1,000,000.\"\n  ...>     }\n  ...> })\n  {:ok,\n      %Instructor.Demos.SpamPrediction{\n          class: :spam\n          score: 0.999\n      }}","ref":"Instructor.html#chat_completion/1-examples"},{"type":"behaviour","title":"Instructor.Adapter","doc":"Behavior for `Instructor.Adapter`.","ref":"Instructor.Adapter.html"},{"type":"callback","title":"Instructor.Adapter.chat_completion/1","doc":"","ref":"Instructor.Adapter.html#c:chat_completion/1"},{"type":"module","title":"Instructor.Adapters.Llamacpp","doc":"Runs against the llama.cpp server. To be clear this calls the llamacpp specific\nendpoints, not the open-ai compliant ones.\n\nYou can read more about it here:\n  https://github.com/ggerganov/llama.cpp/tree/master/examples/server","ref":"Instructor.Adapters.Llamacpp.html"},{"type":"function","title":"Instructor.Adapters.Llamacpp.chat_completion/2","doc":"Run a completion against the llama.cpp server, not the open-ai compliant one.\nThis gives you more specific control over the grammar, and the ability to\nprovide other parameters to the specific LLM invocation.\n\nYou can read more about the parameters here:\n  https://github.com/ggerganov/llama.cpp/tree/master/examples/server","ref":"Instructor.Adapters.Llamacpp.html#chat_completion/2"},{"type":"function","title":"Examples - Instructor.Adapters.Llamacpp.chat_completion/2","doc":"iex> Instructor.chat_completion(%{\n  ...>   model: \"mistral-7b-instruct\",\n  ...>   messages: [\n  ...>     %{ role: \"user\", content: \"Classify the following text: Hello I am a Nigerian prince and I would like to send you money!\" },\n  ...>   ],\n  ...>   response_model: response_model,\n  ...>   temperature: 0.5,\n  ...> })","ref":"Instructor.Adapters.Llamacpp.html#chat_completion/2-examples"},{"type":"module","title":"Instructor.Adapters.OpenAI","doc":"Documentation for `Instructor.Adapters.OpenAI`.","ref":"Instructor.Adapters.OpenAI.html"},{"type":"behaviour","title":"Instructor.EctoType","doc":"","ref":"Instructor.EctoType.html"},{"type":"callback","title":"Instructor.EctoType.to_json_schema/0","doc":"","ref":"Instructor.EctoType.html#c:to_json_schema/0"},{"type":"module","title":"Instructor.JSONSchema","doc":"","ref":"Instructor.JSONSchema.html"},{"type":"function","title":"Instructor.JSONSchema.from_ecto_schema/1","doc":"Generates a JSON Schema from an Ecto schema.\n\n  Note: This will output a correct JSON Schema for the given Ecto schema, but\n  it will not necessarily be optimal, nor support all Ecto types.","ref":"Instructor.JSONSchema.html#from_ecto_schema/1"},{"type":"function","title":"Instructor.JSONSchema.title_for/1","doc":"","ref":"Instructor.JSONSchema.html#title_for/1"},{"type":"module","title":"Instructor.JSONStreamParser","doc":"","ref":"Instructor.JSONStreamParser.html"},{"type":"function","title":"Instructor.JSONStreamParser.parse/1","doc":"","ref":"Instructor.JSONStreamParser.html#parse/1"},{"type":"behaviour","title":"Instructor.Validator","doc":"By default you'll get whatever OpenAI returns.\nThis behavior provides a hook for you to critique the response using standard ecto changesets validations.\nThis can be used in conjuction with the `:max_retries` parameter to `Instructor.chat_completion/1` to retry the completion until it passes your validation.","ref":"Instructor.Validator.html"},{"type":"behaviour","title":"Examples - Instructor.Validator","doc":"defmodule Instructor.Demos.SpamPrediction do\n    use Ecto.Schema\n\n    @primary_key false\n    schema \"spam_prediction\" do\n        field :class, :string\n        field :score, :float\n    end\n\n    @impl true\n    def validate_changeset(changeset) do\n        changeset\n        |> validate_number(:score, less_than_or_equal_to: 1.0)\n    end\n  end\n\n  iex> Instructor.chat_completion(%{\n  ...>   model: \"gpt-3.5-turbo\",\n  ...>   response_model: Instructor.Demos.SpamPrediction,\n  ...>   max_retries: 1,\n  ...>   messages: [\n  ...>     %{\n  ...>       role: \"user\",\n  ...>       content: \"Classify the following text: Hello, I am a Nigerian prince and I would like to give you $1,000,000.\"\n  ...>     }\n  ...> })\n  {:error, %Ecto.Changeset{\n      action: nil,\n      changes: %{},\n      errors: [\n          score: {\"is invalid\", [validation: :number, validation_opts: [less_than_or_equal_to: 1.0]]}\n      ],\n      data: %Instructor.Demos.SpamPrediction{\n          class: nil,\n          score: nil\n      },\n      valid?: false\n  }}","ref":"Instructor.Validator.html#module-examples"},{"type":"callback","title":"Instructor.Validator.validate_changeset/1","doc":"","ref":"Instructor.Validator.html#c:validate_changeset/1"},{"type":"callback","title":"Instructor.Validator.validate_changeset/2","doc":"","ref":"Instructor.Validator.html#c:validate_changeset/2"},{"type":"extras","title":"instructor_ex","doc":"# instructor_ex\n\n_Structured, Ecto outputs with OpenAI (and OSS LLMs)_\n\n---\n\n[![Instructor version](https://img.shields.io/hexpm/v/instructor.svg)](https://hex.pm/packages/instructor)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-lightgreen.svg)](https://hexdocs.pm/instructor/)\n[![Hex Downloads](https://img.shields.io/hexpm/dt/instructor)](https://hex.pm/packages/instructor)\n[![GitHub stars](https://img.shields.io/github/stars/thmsmlr/instructor_ex.svg)](https://github.com/thmsmlr/instructor_ex/stargazers)\n[![Twitter Follow](https://img.shields.io/twitter/follow/thmsmlr?style=social)](https://twitter.com/thmsmlr)\n[![Discord](https://img.shields.io/discord/1192334452110659664?label=discord)](https://discord.gg/CV8sPM5k5Y)\n\n<!-- Docs -->\n\nInstructor.ex is a spiritual port of the great [Instructor Python Library](https://github.com/jxnl/instructor) by [@jxnlco](https://twitter.com/jxnlco), check out his [talk](https://www.youtube.com/watch?v=yj-wSRJwrrc) on youtube.\nThis library brings structured prompting to LLMs. Instead of receiving text as output, Instructor will coax the LLM to output valid JSON that maps directly to the provided Ecto schema.\nIf the LLM fails to do so, or provides values that do not pass your validations, it will provide you utilities to automatically retry with the LLM to correct errors.\nBy default it's designed to be used with the [OpenAI API](https://platform.openai.com/docs/api-reference/chat-completions/create), however it provides an extendable adapter behavior to work with [ggerganov/llama.cpp](https://github.com/ggerganov/llama.cpp) and [Bumblebee (Coming Soon!)](https://github.com/elixir-nx/bumblebee).\n\nAt its simplest, usage is pretty straightforward,\n\n```elixir\ndefmodule SpamPrediction do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @doc \"\"\"","ref":"readme.html"},{"type":"extras","title":"Field Descriptions: - instructor_ex","doc":"- class: Whether or not the email is spam\n  - reason: A short, less than 10 word rationalization for the classification\n  - score: A confidence score between 0.0 and 1.0 for the classification\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl true\n  def validate_changeset(changeset) do\n    changeset\n    |> Ecto.Changeset.validate_number(:score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n\nis_spam? = fn text ->\n  Instructor.chat_completion(\n    model: \"gpt-3.5-turbo\",\n    response_model: SpamPredicition,\n    max_retries: 3,\n    messages: [\n      %{\n        role: \"user\",\n        content: \"\"\"\n        You purpose is to classify customer support emails as either spam or not.\n        This is for a clothing retailer business.\n        They sell all types of clothing.\n\n        Classify the following email: #{text}\n        \"\"\"\n      }\n    ]\n  )\nend\n\nis_spam?.(\"Hello I am a Nigerian prince and I would like to send you money\")\n\n# => {:ok, %SpamPrediction{class: :spam, reason: \"Nigerian prince email scam\", score: 0.98}}\n```\n\nSimply create an ecto schema, optionally provide a `@doc` to the schema definition which we pass down to the LLM, then make a call to `Instructor.chat_completion/1` with context about the task you'd like the LLM to complete.\nYou can also provide a `validate_changeset/1` function via the `use Instructor.Validator` which will provide a set of code level ecto changeset validations. You can use this in conjunction with `max_retries: 3` to automatically, iteratively go back and forth with the LLM up to `n` times with any validation errors so that it has a chance to fix them.\n\n**Curious to learn more? Unsure of how you'd use this? Check out our extensive set of [tutorials](https://hexdocs.pm/instructor/tutorial.html)**\n\n* [Tutorial - Basic Usage & Features](https://hexdocs.pm/instructor/tutorial.html)\n* [Text Classification](https://hexdocs.pm/instructor/text-classification.html)\n* [Extracting Action Items from Meeting Transcriptions](https://hexdocs.pm/instructor/extract-action-items-from-meeting-transcripts.html)\n* [Extracting Explorer.DataFrames from text](https://hexdocs.pm/instructor/text-to-dataframes.html)","ref":"readme.html#field-descriptions"},{"type":"extras","title":"Configuration - instructor_ex","doc":"To configure the default OpenAI adapter you can set the configuration,\n\n```elixir\nconfig :openai, api_key: \"sk-........\"\nconfig :openai, http_options: [recv_timeout: 10 * 60 * 1000]\n```\n\nTo use a local LLM, you can install and run [llama.cpp server](https://github.com/ggerganov/llama.cpp/blob/master/examples/server/README.md) and tell instructor to use it,\n\n```elixir\nconfig :instructor, adapter: Instructor.Adapters.Llamacpp\n```\n\n<!-- Docs -->","ref":"readme.html#configuration"},{"type":"extras","title":"Installation - instructor_ex","doc":"In your mix.exs,\n\n```elixir\ndef deps do\n  [\n    {:instructor, \"~> 0.0.3\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"TODO - instructor_ex","doc":"- [ ] llamacpp adapter broken, needs to support openai input/output API\n  - [ ] GBNF should enforce required properties on objects, currently they're optional.\n  - [ ] GBNF limit the number of digits in number tokens -- small models can sometimes run off to infinit digits\n- [ ] Add `llm_validator`\n- [ ] Logging for Distillation / Finetuning\n- [ ] Add a Bumblebee adapter\n- [ ] Add llamacpp_ex adapter\n- [x] Support Streaming\n  - [ ] Verify schemaless support `{:array, %{name: :string}}`\n  - [ ] Support typespec style support for array streaming `[MySchema]`\n- [ ] Optional/Maybe types\n- [ ] When changeset fails to cast, it should show the underlying data (better error messages)\n- [ ] Add Livebook Tutorials, include in Hexdocs\n    - [x] Text Classification\n    - [ ] Self Critique\n    - [ ] Image Extracting Tables\n    - [ ] Moderation\n    - [x] Citations\n    - [ ] Knowledge Graph\n    - [ ] Entity Resolution\n    - [ ] Search Queries\n    - [ ] Query Decomposition\n    - [ ] Recursive Schemas\n    - [x] Table Extraction\n    - [x] Action Item and Dependency Mapping\n    - [ ] Multi-File Code Generation\n    - [ ] PII Data Sanitizatiommersed\n- [x] Update hexdocs homepage to include example for tutorial\n- [ ] Setup Github CI for testing, add badge to README","ref":"readme.html#todo"},{"type":"extras","title":"Blog Posts - instructor_ex","doc":"- [ ] Why structured prompting?\n\n    Meditations on new HCI.\n    Finally we have software that can understand text. f(text) -> text.\n    This is great, as it gives us a new domain, but the range is still text.\n    While we can use string interpolation to map Software 1.0 into f(text), the outputs are not interoperable with Software 1.0.\n    Hence why UXs available to us are things like Chatbots as our users have to interpret the output.\n\n    Instructor, structure prompting, gives use f(text) -> ecto_schema.\n    Schemas are the lingua franca of Software 1.0.\n    With Instrutor we can now seamlessly move back and forth between Software 1.0 and Software 2.0.\n\n    Now we can maximally leverage AI...\n\n- [ ] From GPT-4 to zero-cost production - Distilation, local-llms, and the cost structure of AI.\n\n    ... 😘","ref":"readme.html#blog-posts"},{"type":"extras","title":"Tutorial","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Tutorial\n\n```elixir\nMix.install(\n  [\n    # {:instructor, \">= 0.0.2\"}\n    {:instructor, path: \"/Users/thomas/code/instructor_ex\"}\n  ],\n  config: [\n    instructor: [\n      adapter: Instructor.Adapters.OpenAI\n    ],\n    openai: [\n      api_key: System.fetch_env!(\"LB_OPENAI_API_KEY\"),\n      http_options: [recv_timeout: 10 * 60 * 1000]\n    ]\n  ]\n)\n```","ref":"tutorial.html"},{"type":"extras","title":"Introduction - Tutorial","doc":"Instructor is a library to do structured prompting with OpenAI and open source LLMs. While the idea is pretty simple, through this and the other examples you'll realize how powerful a concept this is.\n\n**So first off, what is structure prompting?**\n\nWhat if the LLM returned data conforming to a complicated nested schema that your code knows how to work with? Well, that's structure prompting. It's a way of cohercing the LLM to producing it's response in a known format that your downstream code can handle. In the case of Instructor, we use Ecto to provide those schemas. Good old Ecto, something you're already familiar with.\n\nSo, without further ado, let's take define a schema and take it for a spin!\n\n```elixir\ndefmodule Politician do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @doc \"\"\"\n  A description of United States Politicians and the offices that they held,","ref":"tutorial.html#introduction"},{"type":"extras","title":"Fields: - Tutorial","doc":"- first_name: Their first name\n  - last_name: Their last name\n  - offices_held:\n    - office: The branch and position in government they served in\n    - from_date: When they entered office or null\n    - until_date: The date they left office or null\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:first_name, :string)\n    field(:last_name, :string)\n\n    embeds_many :offices_held, Office, primary_key: false do\n      field(:office, Ecto.Enum,\n        values: [:president, :vice_president, :govenor, :congress, :senate]\n      )\n\n      field(:from_date, :date)\n      field(:to_date, :date)\n    end\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Politician, <<70, 79, 82, 49, 0, 0, 17, ...>>,\n [__schema__: 1, __schema__: 1, __schema__: 1, __schema__: 1, __schema__: 2, __schema__: 2, ...]}\n```\n\nGreat, we have our schema describing politicans and the offices they held. Let's notice a few things that may stand out from regular Ecto usage. First, since there is no database backing the schema, it doesn't make sense to give it a primary_key. This also makes sense because there is no sensible value for the LLM to respond with.\n\nAlso we use a `@doc` on the schema. This isn't just for documentation purposes of the tutorial. Instructor will take any `@doc` tag and provide it to the LLM. Generally you'll want to use this to provide semantic descriptions of the fields and general context to the LLM to ensure you get the outputs you want. In our case we want to push the LLM to understand that we are only considering American politicians.\n\nSo, let's try asking the LLM to give us some politicians.\n\n```elixir\nInstructor.chat_completion(\n  model: \"gpt-3.5-turbo\",\n  response_model: Politician,\n  messages: [\n    %{\n      role: \"user\",\n      content:\n        \"Who won the American 2020 election and what offices have they held over their career?\"\n    }\n  ]\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok,\n %Politician{\n   first_name: \"Joe\",\n   last_name: \"Biden\",\n   offices_held: [%Politician.Office{office: :president, from_date: ~D[2021-01-20], to_date: nil}]\n }}\n```\n\nAmazing, right? Using nothing more than one of the top libraries in Elixir, Ecto, we were able to get structured output from our LLM. The data returned is ready to be processed by our regular Elixir code. Instructor supports all field types that you can express in Ecto, including embedded and associated schemas.\n\nIt's almost as if the LLM inputted the data into a Phoenix Form. All the utilities that you use to process that kind of data, you can use to process the outputs of Instructor.\n\nOne of the superpowers of this is that since we're just using changesets under the hood, you can use the same validations that you would use elsewhere in your app. Let's look at that in the next section.","ref":"tutorial.html#fields"},{"type":"extras","title":"Validations - Tutorial","doc":"Instructor provides a lightweight behavior where you can define a callback function that we will call to validate the data returned by the LLM using Ecto changesets. There is nothing fancy to this API. It's just a changeset in and a changeset out.\n\n```elixir\ndefmodule NumberSeries do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @primary_key false\n  embedded_schema do\n    field(:series, {:array, :integer})\n  end\n\n  @impl true\n  def validate_changeset(changeset) do\n    changeset\n    |> Ecto.Changeset.validate_length(:series, min: 10)\n    |> Ecto.Changeset.validate_change(:series, fn\n      field, values ->\n        if Enum.sum(values) |> rem(2) == 0 do\n          []\n        else\n          [{field, \"The sum of the series must be even\"}]\n        end\n    end)\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, NumberSeries, <<70, 79, 82, 49, 0, 0, 18, ...>>, {:validate_changeset, 1}}\n```\n\nIn this albeit contrived example, we're going to get the LLM to return a series of numbers and validate whether it has at least 10 numbers and that the sum of the series is even.\n\nWhen we ask for fewer than ten numbers, Instructor will return an error tuple with a change set that is invalid.\n\n```elixir\n{:error, changeset} =\n  Instructor.chat_completion(\n    model: \"gpt-3.5-turbo\",\n    response_model: NumberSeries,\n    messages: [\n      %{role: \"user\", content: \"Give me the first 5 integers\"}\n    ]\n  )\n\n# Render our the errors down to strings.\nerrors =\n  Ecto.Changeset.traverse_errors(changeset, fn {msg, opts} ->\n    Regex.replace(~r\"%{(\\w+)}\", msg, fn _, key ->\n      opts |> Keyword.get(String.to_existing_atom(key), key) |> to_string()\n    end)\n  end)\n\n{changeset.changes, errors}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{%{series: [1, 2, 3, 4, 5]},\n %{series: [\"The sum of the series must be even\", \"should have at least 10 item(s)\"]}}\n```\n\nNow the beauty of this is that since we have human readable errors from our validations, we can just turn around and pass those back into the LLM to get it to fix its own errors.\n\nInstructor provides a convenience parameter, `max_retries` for you in the initial call which will retry against the validations up to n times.\n\n```elixir\nInstructor.chat_completion(\n  model: \"gpt-3.5-turbo\",\n  response_model: NumberSeries,\n  max_retries: 10,\n  messages: [\n    %{role: \"user\", content: \"Give some random integers\"}\n  ]\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n\n23:59:13.641 [debug] Retrying LLM call for Elixir.NumberSeries...\n\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok, %NumberSeries{series: [2, 4, 6, 8, 10, 12, 14, 16, 18, 20]}}\n```\n\n<!-- livebook:{\"offset\":6729,\"stamp\":{\"token\":\"XCP.Roo6JXuq5pvsYYM4xHSSXzUQNOfPTuMS0cQOeSVWO-9Am1Nv76KPHbWYwzCJDEP0xm_Bl7ignzLNlkI1l5sOtwmp5PcUrye8NltUPM1MOwiAhT_ZRLKKk9w\",\"version\":2}} -->","ref":"tutorial.html#validations"},{"type":"extras","title":"Text Classification","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Text Classification\n\n```elixir\nMix.install(\n  [\n    {:instructor, \"~> 0.0.2\"}\n  ],\n  config: [\n    instructor: [\n      adapter: Instructor.Adapters.OpenAI\n    ],\n    openai: [\n      api_key: System.fetch_env!(\"LB_OPENAI_API_KEY\"),\n      http_options: [recv_timeout: 10 * 60 * 1000]\n    ]\n  ]\n)\n```","ref":"text-classification.html"},{"type":"extras","title":"Motivation - Text Classification","doc":"Text classification is a common task in NLP and broadly applicable across software. Whether it be spam detection, or support ticket categorization, NLP is at the core. Historically, this required training custom, bespoke models that required collecting thousands of pre-labeled examples. With LLMs a lot of this knowledge is already encoded into the model. With proper instruction and guiding the output to a known set of classifications using GPT you can be up and running with a text classification model in no time.\n\nHell, you can even use instructor to help generate the training set to train your own more efficient model. But let's not get ahead of ourselves, there's more on that later in the tutorials.","ref":"text-classification.html#motivation"},{"type":"extras","title":"Binary Text Classification - Text Classification","doc":"Spam detection is a classic example of binary text classification. It's as simple as returning a true / false of whether an example is in the class. This is pretty trivial to implement in instructor.\n\n```elixir\ndefmodule SpamClassification do\n  use Ecto.Schema\n\n  @doc \"\"\"\n  A classification of whether or not a provided email is spam or not\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:is_spam?, :boolean)\n  end\nend\n\nis_spam? = fn text ->\n  {:ok, %{is_spam?: result}} =\n    Instructor.chat_completion(\n      model: \"gpt-3.5-turbo\",\n      response_model: SpamClassification,\n      messages: [\n        %{\n          role: \"user\",\n          content: \"Classify the following text as spam/not_spam: #{text}\"\n        }\n      ]\n    )\n\n  result\nend\n\nis_spam?.(\"Hello I am a Nigerian prince and I would like to send you money\")\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\ntrue\n```","ref":"text-classification.html#binary-text-classification"},{"type":"extras","title":"Multi-Class Text Classification - Text Classification","doc":"We don't have to stop just at a boolean inclusion, we can also easily extend this idea to multiple categories or classes that we can classify the text into. In this example, let's consider classifying support emails. We want to know whether it's a `general_inquiry`, `billing_issue`, or a `technical_issue` perhaps it rightly fits in multiple classes. This can be useful if we want to cc' specialized support agents when intersecting customer issues occur\n\nWe can leverage `Ecto.Enum` to define a schema that restricts the LLM output to be a list one of those values. We can also provide a `@doc` description to help guide the LLM with the semantic understanding of what these classifications ought to represent.\n\n```elixir\ndefmodule EmailClassifications do\n  use Ecto.Schema\n\n  @doc \"\"\"\n  A classification of a customer support email.\n\n  technical_issue - whether the user is having trouble accessing their account.\n  billing_issue - whether the customer is having trouble managing their billing or credit card\n  general_inquiry - all other issues\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:tags, {:array, Ecto.Enum},\n      values: [:general_inquiry, :billing_issue, :technical_issue]\n    )\n  end\nend\n\nclassify_email = fn text ->\n  {:ok, %{tags: result}} =\n    Instructor.chat_completion(\n      model: \"gpt-3.5-turbo\",\n      response_model: EmailClassifications,\n      messages: [\n        %{\n          role: \"user\",\n          content: \"Classify the following text: #{text}\"\n        }\n      ]\n    )\n\n  result\nend\n\nclassify_email.(\"My account is locked and I can't access my billing info.\")\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[:technical_issue, :billing_issue]\n```\n\n<!-- livebook:{\"offset\":3715,\"stamp\":{\"token\":\"XCP.5TEM1YsHThC1nSHs6aMACbc9nSINfaTw-AQHdMj8mhASbpSXy0vET32EAGwwrjDer0tOheln-yInCPR3IuLXR8U8Ytq84j9lNIhF9sQAafqSS532tBie-4E\",\"version\":2}} -->","ref":"text-classification.html#multi-class-text-classification"},{"type":"extras","title":"Extracting Action Items from Meeting Transcripts","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Extracting Action Items from Meeting Transcripts\n\n```elixir\nMix.install(\n  [\n    # {:instructor, \"~> 0.0.2\"}\n    {:instructor, path: \"/Users/thomas/code/instructor_ex\"},\n    {:kino, \"~> 0.12.0\"}\n  ],\n  config: [\n    instructor: [\n      adapter: Instructor.Adapters.OpenAI\n    ],\n    openai: [\n      api_key: System.fetch_env!(\"LB_OPENAI_API_KEY\"),\n      http_options: [recv_timeout: 10 * 60 * 1000]\n    ]\n  ]\n)\n```","ref":"extract-action-items-from-meeting-transcripts.html"},{"type":"extras","title":"Motivation - Extracting Action Items from Meeting Transcripts","doc":"This example shouldn't be foreign to any of you. I'm sure each day you log in, join a Zoom call, and have a meeting about what y'all are going to do this upcoming week. It is then the work of some product manager to translate this into a JIRA board so that you can track your progress throughout the week.\n\nThe bane of most engineers existence... We can automate this.","ref":"extract-action-items-from-meeting-transcripts.html#motivation"},{"type":"extras","title":"The Schema - Extracting Action Items from Meeting Transcripts","doc":"Let's start by defining a schema for the tickets and the subtasks that might exist within them. There will be a priority, description and a set of dependencies between the tasks for the tickets.\n\n```elixir\ndefmodule MeetingNotes do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @doc \"\"\"\n  Tickets correctly resolved from a meeting transcription.\n\n  Tickets have a name, priority, useful description, and assignees.\n  They may also have subtasks that share the same id spaces as the tickets.\n  A ticket also may have references to Tickets and Subtasks that are a blocking dependency\n  for completing the ticket.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    embeds_many :tickets, Ticket do\n      field(:name, :string)\n      field(:description, :string)\n      field(:priority, Ecto.Enum, values: [:high, :medium, :low])\n      field(:assignees, {:array, :string})\n\n      embeds_many :subtasks, SubTasks do\n        field(:name, :string)\n      end\n\n      field(:dependencies, {:array, :binary_id})\n    end\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, MeetingNotes, <<70, 79, 82, 49, 0, 0, 17, ...>>,\n [__schema__: 1, __schema__: 1, __schema__: 1, __schema__: 1, __schema__: 2, __schema__: 2, ...]}\n```\n\nOne thing interesting to note here is that although we could use proptorecto associations, in this case it's rather tedious. Since our dependencies can either be on tickets or on subtasks, you would typically have to have a polymorphic relationship.\n\nWhile Instructor fully supports doing that, we can instead just create embedded schemas for the subtasks and define that the subtasks and the tickets share the same ID space in our doc comment. This will steer the LLM to produce foreign key relations in a polymorphic way.\n\nOne also might choose to solve this modeling problem by denormalizing the tickets and subtasks into its own task type and get the LLM to not only have the dependencies listed but also any subtasks that a ticket might have through association instead of embedding.\n\nEither way works, but in this example we prefer the embedded method because it produces a prettier output and we can lean on the LLM a little more.\n\nNow, let's extract some JIRA tickets.\n\n```elixir\ngenerate_tickets = fn transcript ->\n  Instructor.chat_completion(\n    model: \"gpt-3.5-turbo\",\n    response_model: MeetingNotes,\n    messages: [\n      %{role: \"system\", content: \"The following is a transcript of a meeting...\"},\n      %{\n        role: \"user\",\n        content: \"Create the action items for the following transcript: #{transcript}\"\n      }\n    ]\n  )\nend\n\n{:ok, %{tickets: tickets}} =\n  generate_tickets.(\"\"\"\n    Alice: Hey team, we have several critical tasks we need to tackle for the upcoming release. First, we need to work on improving the authentication system. It's a top priority.\n    \n    Bob: Got it, Alice. I can take the lead on the authentication improvements. Are there any specific areas you want me to focus on?\n    \n    Alice: Good question, Bob. We need both a front-end revamp and back-end optimization. So basically, two sub-tasks.\n    \n    Carol: I can help with the front-end part of the authentication system.\n    \n    Bob: Great, Carol. I'll handle the back-end optimization then.\n    \n    Alice: Perfect. Now, after the authentication system is improved, we have to integrate it with our new billing system. That's a medium priority task.\n    \n    Carol: Is the new billing system already in place?\n    \n    Alice: No, it's actually another task. So it's a dependency for the integration task. Bob, can you also handle the billing system?\n    \n    Bob: Sure, but I'll need to complete the back-end optimization of the authentication system first, so it's dependent on that.\n    \n    Alice: Understood. Lastly, we also need to update our user documentation to reflect all these changes. It's a low-priority task but still important.\n    \n    Carol: I can take that on once the front-end changes for the authentication system are done. So, it would be dependent on that.\n    \n    Alice: Sounds like a plan. Let's get these tasks modeled out and get started.\n  \"\"\")\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok,\n %MeetingNotes{\n   tickets: [\n     %MeetingNotes.Ticket{\n       id: \"T1\",\n       name: \"Improving authentication system\",\n       description: \"Work on improving the authentication system\",\n       priority: :high,\n       assignees: [\"Bob\"],\n       subtasks: [\n         %MeetingNotes.Ticket.SubTasks{id: \"T1.1\", name: \"Front-end revamp\"},\n         %MeetingNotes.Ticket.SubTasks{id: \"T1.2\", name: \"Back-end optimization\"}\n       ],\n       dependencies: []\n     },\n     %MeetingNotes.Ticket{\n       id: \"T2\",\n       name: \"Integrating authentication system with billing system\",\n       description: \"Integrate the improved authentication system with the new billing system\",\n       priority: :medium,\n       assignees: [\"Bob\"],\n       subtasks: [],\n       dependencies: [\"T1.2\"]\n     },\n     %MeetingNotes.Ticket{\n       id: \"T3\",\n       name: \"Update user documentation\",\n       description: \"Update user documentation to reflect the changes\",\n       priority: :low,\n       assignees: [\"Carol\"],\n       subtasks: [],\n       dependencies: [\"T1.1\"]\n     },\n     %MeetingNotes.Ticket{\n       id: \"T4\",\n       name: \"Handle front-end part of authentication system\",\n       description: \"Help with the front-end part of the authentication system\",\n       priority: :medium,\n       assignees: [\"Carol\"],\n       subtasks: [],\n       dependencies: []\n     },\n     %MeetingNotes.Ticket{\n       id: \"T5\",\n       name: \"Handle back-end optimization of authentication system\",\n       description: \"Handle the back-end optimization of the authentication system\",\n       priority: :high,\n       assignees: [\"Bob\"],\n       subtasks: [],\n       dependencies: []\n     },\n     %MeetingNotes.Ticket{\n       id: \"T6\",\n       name: \"Handle billing system\",\n       description: \"Handle the new billing system\",\n       priority: :medium,\n       assignees: [\"Bob\"],\n       subtasks: [],\n       dependencies: [\"T5\"]\n     }\n   ]\n }}\n```\n\nThe results look good. We can use Kino to then render out all the dependencies of our tickets.\n\n```elixir\ngenerate_tickets_diagram = fn tickets ->\n  subtasks = Enum.flat_map(tickets, & &1.subtasks)\n  all_tasks = subtasks ++ tickets\n\n  ticket_nodes =\n    tickets\n    |> Enum.map_join(\"\\n\", fn t ->\n      subtask_relations =\n        t.subtasks\n        |> Enum.map_join(\"\\n\", fn st ->\n          \"\"\"\n          \"#{t.name}\" ||--o| \"#{st.name}\" : \"Has Subtask\"\n          \"\"\"\n        end)\n\n      dependency_relations =\n        t.dependencies\n        |> Enum.map_join(\"\\n\", fn d ->\n          dt = all_tasks |> Enum.find(&(&1.id == d))\n\n          if dt do\n            \"\"\"\n            \"#{t.name}\" ||--o| \"#{dt.name}\" : \"Depends on\"\n            \"\"\"\n          else\n            \"\"\n          end\n        end)\n\n      \"\"\"\n        \"#{t.name}\" {\n          priority #{t.priority}\n          assignees #{Enum.join(t.assignees, \", \")}\n        }\n\n        #{subtask_relations}\n        #{dependency_relations}\n      \"\"\"\n    end)\n\n  subtask_nodes =\n    tickets\n    |> Enum.flat_map(& &1.subtasks)\n    |> Enum.map_join(\"\\n\", fn st ->\n      \"\"\"\n        \"#{st.name}\"\n      \"\"\"\n    end)\n\n  Kino.Mermaid.new(\"\"\"\n    erDiagram\n      #{ticket_nodes}\n      #{subtask_nodes}\n  \"\"\")\nend\n\ngenerate_tickets_diagram.(tickets)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```mermaid\n  erDiagram\n      \"Improving authentication system\" {\n    priority high\n    assignees Bob\n  }\n\n  \"Improving authentication system\" ||--o| \"Front-end revamp\" : \"Has Subtask\"\n\n\"Improving authentication system\" ||--o| \"Back-end optimization\" : \"Has Subtask\"\n\n  \n\n  \"Integrating authentication system with billing system\" {\n    priority medium\n    assignees Bob\n  }\n\n  \n  \"Integrating authentication system with billing system\" ||--o| \"Back-end optimization\" : \"Depends on\"\n\n\n  \"Update user documentation\" {\n    priority low\n    assignees Carol\n  }\n\n  \n  \"Update user documentation\" ||--o| \"Front-end revamp\" : \"Depends on\"\n\n\n  \"Handle front-end part of authentication system\" {\n    priority medium\n    assignees Carol\n  }\n\n  \n  \n\n  \"Handle back-end optimization of authentication system\" {\n    priority high\n    assignees Bob\n  }\n\n  \n  \n\n  \"Handle billing system\" {\n    priority medium\n    assignees Bob\n  }\n\n  \n  \"Handle billing system\" ||--o| \"Handle back-end optimization of authentication system\" : \"Depends on\"\n\n\n      \"Front-end revamp\"\n\n  \"Back-end optimization\"\n\n\n```\n\n<!-- livebook:{\"offset\":9344,\"stamp\":{\"token\":\"XCP.yoxkI3zUVgtejiyU2srLCvVLT75QXTgesEQjLStJbAnsnGXKP_GDcj9sBvH0HWHPyCeWv5ED8A9E9U3pTmkv52a7CDbpnRt67H0VohcnXKo9I9ftFkEfoC8\",\"version\":2}} -->","ref":"extract-action-items-from-meeting-transcripts.html#the-schema"},{"type":"extras","title":"Converting Text into Explorer DataFrames","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Converting Text into Explorer DataFrames\n\n```elixir\nMix.install(\n  [\n    # {:instructor, \"~> 0.0.2\"}\n    {:instructor, path: \"/Users/thomas/code/instructor_ex\"},\n    {:explorer, \"~> 0.7.2\"},\n    {:kino, \"~> 0.12.0\"},\n    {:kino_explorer, \"~> 0.1.13\"}\n  ],\n  config: [\n    instructor: [\n      adapter: Instructor.Adapters.OpenAI\n    ],\n    openai: [\n      api_key: System.fetch_env!(\"LB_OPENAI_API_KEY\"),\n      http_options: [recv_timeout: 10 * 60 * 1000]\n    ]\n  ]\n)\n```","ref":"text-to-dataframes.html"},{"type":"extras","title":"Motivation - Converting Text into Explorer DataFrames","doc":"Sometimes we come across text that can be naturally extracted into structured data. Although we don't necessarily know how many tables, and what the schemas are represented in the data. Using Instructor we can let the LLM determine the tables and schemas for us, pulling the results into a dataframe which we can then use for further analysis.","ref":"text-to-dataframes.html#motivation"},{"type":"extras","title":"The Schema - Converting Text into Explorer DataFrames","doc":"Notice from this example we want resulting data structure to have data frames. That's not a default Ecto type so we're gonna have to do things a little different here.\n\nLuckily, Ecto allows us to define our own custom types in which we can override how the values are casted. And Instructor provides us a behavior which tells us how to represent it in a JSONSchema so the LLM can understand it.\n\nIn our case, we're going to create a type for a table of data, which will be represented as a CSV string with respect to the LLM. Our custom type will then parse the string when it is casted by Ecto and parse the CSV into an `Explorer.DataFrame`.\n\n```elixir\ndefmodule Ecto.CSVDataFrame do\n  use Ecto.Type\n  use Instructor.EctoType\n\n  def type, do: :string\n\n  def cast(csv_str) when is_binary(csv_str) do\n    df = Explorer.DataFrame.load_csv!(csv_str)\n    {:ok, df}\n  end\n\n  def cast(%Explorer.DataFrame{} = df), do: {:ok, df}\n  def cast(_), do: :error\n\n  def to_json_schema(),\n    do: %{\n      type: \"string\",\n      description: \"A CSV representation of a data table\"\n    }\n\n  def dump(x), do: {:ok, x}\n  def load(x), do: {:ok, x}\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Ecto.CSVDataFrame, <<70, 79, 82, 49, 0, 0, 12, ...>>, {:load, 1}}\n```\n\nNow that we have our data type, we can use it in our embedded ecto schema, just as we would have in any other example. If everything works correctly, we should get some data frames out.\n\n```elixir\ndefmodule Database do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @doc \"\"\"\n  The extracted database will contain one or more tables with data as csv formatted with ',' delimiters\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    embeds_many :tables, DataFrame, primary_key: false do\n      field(:name, :string)\n      field(:data, Ecto.CSVDataFrame)\n    end\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Database, <<70, 79, 82, 49, 0, 0, 16, ...>>,\n [__schema__: 1, __schema__: 1, __schema__: 1, __schema__: 1, __schema__: 2, __schema__: 2, ...]}\n```\n\nPerfect. And now let's test it out on some example text.\n\n```elixir\nextract_df = fn text ->\n  Instructor.chat_completion(\n    model: \"gpt-3.5-turbo\",\n    response_model: Database,\n    messages: [\n      %{\n        role: \"system\",\n        content: \"\"\"\n        Map this data into one or more dataframes and correctly define the columns and data\n        \"\"\"\n      },\n      %{\n        role: \"user\",\n        content: \"#{text}\"\n      }\n    ]\n  )\nend\n\n{:ok, db} =\n  extract_df.(\"\"\"\n    My name is John and I am 25 years old. I live in\n    New York and I like to play basketball. His name is\n    Mike and he is 30 years old. He lives in San Francisco\n    and he likes to play baseball. Sarah is 20 years old\n    and she lives in Los Angeles. She likes to play tennis.\n    Her name is Mary and she is 35 years old.\n    She lives in Chicago.\n    \n    On one team 'Tigers' the captain is John and there are 12 players.\n    On the other team 'Lions' the captain is Mike and there are 10 players.\n  \"\"\")\n\nKino.Layout.tabs(\n  for table <- db.tables do\n    {table.name, table.data}\n  end\n)\n```\n\n<!-- livebook:{\"offset\":3999,\"stamp\":{\"token\":\"XCP.eOIgnqRdBf7UJbu8VBayXk_IAWAeXsVWDLci5cYgC05okdcozpMo6Ws9qnDJgn7gy_mHCuWF4_47W5ZS8-aNkKlULC1M2IO_ZU-D9Dr6c8He4Np36gRxwNc\",\"version\":2}} -->","ref":"text-to-dataframes.html#the-schema"}],"content_type":"text/markdown"}